//  ChainBuilderDW.m
//  KnightsAndSpies

// based on half-remembered Dave Johnson method of building up 2 chains, then 4 chains, etc.
#import "ChainBuilderDW.h"


@implementation ChainBuilderDW
-(ChainBuilderDW*) initWithDW: (Crowd*) givenCrowd
{
    self = [super init];
    crowd = givenCrowd; //Not copied
    return self;
}

//returns array of taggedChains made by joining up knight and spy chains as much as
//possible and the finalKnight
-(NSDictionary*) buildChains
{
    //NSLog(@"buildChains here");
    NSMutableArray* allSpyChains = [[NSMutableArray alloc] init]; //in new scheme each entry is an array of chains
    NSMutableArray* allKnightChains = [[NSMutableArray alloc] init];
    
    NSMutableArray* peopleToConsider = [[NSMutableArray alloc] init];
    int i; for (i=0;i<[crowd numberOfPeople];i++) {
        [peopleToConsider addObject: [NSNumber numberWithInt: i]];
    }

    BOOL firstStep = TRUE;
    while ([peopleToConsider count] > 1) {
        //NSLog(@"peopleToConsider = %@",peopleToConsider);
        NSDictionary* chains = [self buildChainsConsidering: peopleToConsider];
        NSMutableArray* spyChains = [chains objectForKey: @"spyChains"];
        NSMutableArray* knightChains = [chains objectForKey: @"knightChains"];
        NSNumber* singleton = [chains objectForKey: @"singleton"];

        //NSLog(@"singleton = %@",singleton);
        //NSLog(@"spyChains = %@",spyChains);
        //NSLog(@"knightChains = %@",knightChains);
        

        [peopleToConsider release];
        peopleToConsider = [self finalPeople: knightChains];
        if ([singleton intValue] != -1) {
            //NSLog(@"Have a singleton, %i",[singleton intValue]);
            if ([peopleToConsider count] % 2 == 0) {
            //we want the singleton as there are an even number of people (could be 1/2 and 1/2)
            //otherwise there is an absolute majority of knights 
                //NSLog(@"Adding the singleton %i to peopleToConsider",[singleton intValue]);
                [peopleToConsider addObject: singleton];
            }
            if (firstStep) {
                //ONLY PUT in if first step, owise get (a b), (b) and we prefer to resolve
                //such chains by asking about a first
             //always put in the chain (this won't create redundant question as we
            //work backwards)
            //we have a singleton, and have to record its presence
                TaggedChain* singletonChain = [[TaggedChain alloc] init];
                [[singletonChain chain] addObject: singleton];
                [singletonChain setEndsWithSpy: FALSE];
                [knightChains addObject: singletonChain];
            //NSLog(@"knightChains = %@",knightChains);
            }
        }
        else {
            //NSLog(@"No singleton");
        }

        [allKnightChains addObject: knightChains];
        [allSpyChains addObject: spyChains];
        //NSLog(@"peopleToConsider = %@",peopleToConsider);
        firstStep = FALSE;
    }

    //NSLog(@"DW Chainbuilder here");
    //NSLog(@"ready to build results");
    //NSLog(@"allChains = %@",allChains);
    //NSLog(@"finalKnight = %@",[peopleToConsider lastObject]);

    //note nil is essential
    //NSLog(@"results = %@",results);
    //NSLog(@"After forming spyChains and knightChains, results = %@",results);
    NSMutableArray* allChains = [[NSMutableArray alloc] init];
    int j; for (j=0;j<[allKnightChains count];j++) {
        NSMutableArray* currentChains = [[NSMutableArray alloc] init];
        [currentChains addObjectsFromArray: [allKnightChains objectAtIndex: j]];
        [currentChains addObjectsFromArray: [allSpyChains objectAtIndex: j]];
        [allChains addObject: currentChains];
    };
    //NSLog(@"DWbuilder, allChains = %@",allChains);

    NSDictionary* results = [NSDictionary dictionaryWithObjectsAndKeys: allKnightChains, @"knightChains",
        allSpyChains, @"spyChains", allChains, @"allChains",
        [peopleToConsider lastObject], @"finalKnight",nil];
    return results;
}

//returns spyChains and knightChains
-(NSDictionary*) buildChainsConsidering: (NSMutableArray*) peopleToConsider {
    //NSLog(@"buildChainsConsidering: %@",peopleToConsider);
    NSMutableArray* spyChains = [[NSMutableArray alloc] init];
    NSMutableArray* knightChains = [[NSMutableArray alloc] init];
    int n = [peopleToConsider count];
    int i; for (i=0;i<n-1;i+=2) {
        //NSLog(@"i = %i",i);
        int firstPerson = [[peopleToConsider objectAtIndex: i] intValue];
        int secondPerson = [[peopleToConsider objectAtIndex: i+1] intValue];
        TaggedChain* newTaggedChain = [[TaggedChain alloc] init];
        NSMutableArray* newChain = [newTaggedChain chain];
        [newChain addObject: [NSNumber numberWithInt: firstPerson]];
        [newChain addObject: [NSNumber numberWithInt: secondPerson]];
        //NSLog(@"newChain = %@",newChain);
        BOOL shortChain = ([[[crowd person: firstPerson] viewOnPerson: secondPerson] intValue] == SPY);
        if (shortChain) {
            [newTaggedChain setEndsWithSpy: TRUE];
            //NSLog(@"adding to spyChains");
            [spyChains addObject: newTaggedChain];
        }
        else {
            [newTaggedChain setEndsWithSpy: FALSE];
            //NSLog(@"adding to knightChains");
            [knightChains addObject: newTaggedChain];
        }
    }
    NSNumber* singleton;
    if (i == n-1) {  //happens when n is odd, e.g. n=7, i=0, i=2, i=4, i=6 (no iteration on this)
        singleton = [peopleToConsider objectAtIndex: i];
    }
    else {
        singleton = [NSNumber numberWithInt: -1];
    }
    // NSLog(@"spyChains = %@",spyChains);
    // NSLog(@"knightChains = %@",knightChains);
    //NSLog(@"results = %@",results);
    NSDictionary* results = [NSDictionary dictionaryWithObjectsAndKeys: spyChains, @"spyChains",
        knightChains, @"knightChains", singleton, @"singleton", nil]; 
    return results;

}
    
-(NSMutableArray*) finalPeople: (NSMutableArray*) chains //this is now an array of TaggedChains
{
    NSMutableArray* finalPeople = [[NSMutableArray alloc] init];
    int i; for (i=0;i<[chains count];i++) {
        [finalPeople addObject: [[[chains objectAtIndex: i] chain] lastObject]];
    }
    return finalPeople;
}

//This should now be replaced with the generic chainJoiner method
-(NSMutableArray*) joinSpyChains: (NSMutableArray*) spyChains 
                    toKnightChains: (NSMutableArray*) knightChains
{
    NSMutableArray* results = [[NSMutableArray alloc] init];
    NSMutableArray* spyChainsCopy = [spyChains mutableCopy];
    NSMutableArray* knightChainsCopy = [knightChains mutableCopy];
    
    //we assume that each knight chain can be completed by joining it
    //to chains appearing in the right order later in the knight array
    //and then (possibly) to a spy chain
    int n = [knightChainsCopy count]; //we delete from knightChains, and adjust n accordingly
    int i; for (i=0;i<n;i++) {
        NSMutableArray* root = [[knightChainsCopy objectAtIndex: i] mutableCopy];
        //NSLog(@"Started with root = %@",root);
        int lastMember = [[root lastObject] intValue];
        int j; for (j=i+1;j<n;j++) {
            NSMutableArray* candidateChain = [knightChainsCopy objectAtIndex: j];
            if ([[candidateChain objectAtIndex: 0] intValue] == lastMember) {
                if ([candidateChain count] > 1)
                    [root addObject: [candidateChain lastObject]];
                [knightChainsCopy removeObjectAtIndex: j]; n-=1;
                lastMember = [[root lastObject] intValue];
                //NSLog(@"root = %@",root);
                //NSLog(@"lastMember = %i",lastMember);
            }
        }
        //the chain root is now complete, except possible for a spyChain ending
        BOOL endsInSpy = FALSE;
        int m = [spyChainsCopy count];
        int k; for (k=0;k<m;k++) {
            NSMutableArray* candidateChain = [spyChainsCopy objectAtIndex: k];
            if ([[candidateChain objectAtIndex: 0] intValue] == lastMember) {
                [root addObject: [candidateChain lastObject]];
                //NSLog(@"removing the following chain from spyChains %@",candidateChain);
                [spyChainsCopy removeObjectAtIndex: k]; m-=1;
                endsInSpy = TRUE;
                //allow loop to continue, but won't have two spy chains
            }
        }
        TaggedChain *taggedChain = [[TaggedChain alloc] init];
        [taggedChain setChain: root];
        if (endsInSpy) {
            [taggedChain setEndsWithSpy: TRUE];
            [taggedChain setWeight: [root count]-1];
        }
        else {
            [taggedChain setEndsWithSpy: FALSE];
            [taggedChain setWeight: [root count]];
        }
        [results addObject: taggedChain];
    }
    //finally need to add spyChains left
    int l; for (l=0;l<[spyChainsCopy count];l++) {
        TaggedChain *taggedChain = [[TaggedChain alloc] init];
        [taggedChain setChain: [spyChainsCopy objectAtIndex: l]];
        [taggedChain setEndsWithSpy: TRUE];
        [results addObject: taggedChain];
    }
    
    return results;
}
                
                

@end
