//  ChainResolver.m
//  KnightsAndSpies


#import "ChainResolver.h"

int UNKNOWN = -1;

@implementation ChainResolver

-(ChainResolver*) initWith: (Crowd*) givenCrowd
                 allChains: (NSMutableArray*) givenChains
                    knight: (Person*) givenKnight;
{
    self = [super init];
    crowd = givenCrowd; //Not copied
    //NSLog(@"ChainResolver init here, crowd = %@",crowd);
    //[crowd showMembers];
    allChains = givenChains;
    knight = givenKnight;
    identities = [[NSMutableArray alloc] init];
    int i; for (i=0;i<[crowd numberOfPeople];i++) {
        [identities addObject: [NSNumber numberWithInt: UNKNOWN]];
    }
    //NSLog(@"ChainResolver init here, just made identities = %@",identities);
    return self;
}

-(NSMutableArray*) identities
{
    return identities;
}


-(NSMutableArray*) resolveChains
{
    //NSLog(@"resolveChains here");
    int numberOfKnight = [knight number];
    [identities replaceObjectAtIndex: numberOfKnight
                          withObject: [NSNumber numberWithInt: KNIGHT]];
    //NSLog(@"identities = %@",identities);
    
	//NSLog(@"resolveChains here, allChains = %@",allChains);
	
    int i; for (i=[allChains count]-1;i>=0;i--) { //start with the most crosslinked chains, as they can knock on more
        NSMutableArray* chains = [allChains objectAtIndex: i];
        [self addToIdentitiesUsingChains: (NSMutableArray*) chains
                           resolvingEnds: (i==0) ? TRUE : FALSE];
    }
    return identities;
}

-(NSMutableArray*) resolveChainsAll
{
    //NSLog(@"resolveChains here");
    int numberOfKnight = [knight number];
    [identities replaceObjectAtIndex: numberOfKnight
                          withObject: [NSNumber numberWithInt: KNIGHT]];
    //NSLog(@"identities = %@",identities);
	//NSLog(@"allChains = %@",allChains);

    int i; for (i=[allChains count]-1;i>=0;i--) {
        NSMutableArray* chains = [allChains objectAtIndex: i];
        [self addToIdentitiesUsingChains: (NSMutableArray*) chains
                           resolvingEnds: TRUE];
    }
    return identities;
}


-(void) addToIdentitiesUsingChains: (NSMutableArray*) chains
                     resolvingEnds: (BOOL) resolvingEnds
{
	//NSLog(@"addToIdentitiesUsingChains %@, resolvingEnds = %i",chains,resolvingEnds);
    int i; for (i=0;i<[chains count];i++) {
        TaggedChain* taggedChain = [chains objectAtIndex: i];
        NSMutableArray* chain = [taggedChain chain];
        //NSLog(@"Working with the taggedChain %@",taggedChain);

        int firstKnight = [self determineFirstKnightInTaggedChain: taggedChain
                                           persistWithShortChains: TRUE]; //TRUE replaced resolvingEnds
		//NSLog(@"firstKnight = %i",firstKnight);
        //this function will trim the chain using the identities, then call the existing
        //function for finding a knight, then update the firstKnight instance variable
        //of taggedChain. Should return the chain count - 1 if no knight is found. NOTE
        //this does not mean that the last person is necessarily a knight. If after
        //trimming the chain has length 1 or 2 and we're not resolvingEnds, return -1

        //NSLog(@"firstKnight reported to be in position %i",firstKnight);
        //NSLog(@"after calling function, chain = %@",chain);

        if (firstKnight == -1) {
            //we should still see if we know about the penultimate member
            NSMutableArray* chain = [taggedChain chain];
			//NSLog(@"ChainResolver here, looking at taggedChain %@ with firstknight = -1",taggedChain);
            if ([chain count] > 1) {
                int number = [[chain objectAtIndex: [chain count]-2] intValue];
                if ([[identities objectAtIndex: number] intValue] == KNIGHT) {
                    if ([taggedChain endsWithSpy])
                    {
                        [self setPerson: [[chain lastObject] intValue] to: SPY];
                        //NSLog(@"setting last to spy");
                    }
                    else {
                        [self setPerson: [[chain lastObject] intValue] to: KNIGHT];
                        //NSLog(@"setting last to knight");
                    }
                }
            }
            //NSLog(@"Not persisting with this chain");
        }
        else {
            [taggedChain setFirstKnight: firstKnight]; //chain count-1 if no knights
			//NSLog(@"ChainResolver here, looking at chain %@ with firstKnight = %i",chain,firstKnight);

            int j; for (j=0;j<firstKnight;j++) { //e.g. a chain S S ? will have firstKnight = 2, and get S S
                [self setPerson: [[chain objectAtIndex: j] intValue] to: SPY];
                //[identities replaceObjectAtIndex: [[chain objectAtIndex: j] intValue]
                //                      withObject: [NSNumber numberWithInt: SPY]];
                //NSLog(@"Made person %i a spy",[[chain objectAtIndex: j] intValue]);
            }
            int k; for (k=firstKnight;k<[chain count]-1;k++) {
                //go from firstKnight to the penultimate position in the chain
                //so e.g. if chain is S S S K S then firstKnight = 3, and k = 3 < 4 is only iteration
                //if chain is S S ?(said to be S) then firstKnight = 2 then nothing happens
                [self setPerson: [[chain objectAtIndex: k] intValue] to: KNIGHT];
     
                //[identities replaceObjectAtIndex: [[chain objectAtIndex: k] intValue]
                //                      withObject: [NSNumber numberWithInt: KNIGHT]];
                //NSLog(@"Made person %i a knight",[[chain objectAtIndex: k] intValue]);
            }
            if (firstKnight == [chain count]-1) {
                if (resolvingEnds) {
                    //this means there was no knight in the main part of the chain
                    //and we need to ask about the last chap
                    int numberOfLastChap = [[chain lastObject] intValue];
                    NSNumber* lastIdentity;
                    if ([[identities objectAtIndex: numberOfLastChap] intValue] != -1) {
                        lastIdentity = [identities objectAtIndex: numberOfLastChap];
                    }
                    else {
                        lastIdentity = [knight viewOnPerson: numberOfLastChap];
                    }
                    [self setPerson: numberOfLastChap to: [lastIdentity intValue]]; //had wrongly asked again
                    //[identities replaceObjectAtIndex: numberOfLastChap
                    //                      withObject: lastIdentity];
                    //NSLog(@"resolving the end of a chain of the form S ... S ?");
                }
                else {
                    //NSLog(@"Chain of form S ... S ? found, not resolving end");
                }
                //if we're not resolvingEnds, don't bother doing anything here
            }
            else {
                //now we know the chain ends in what ever we thought it ended with
                if ([taggedChain endsWithSpy])
                {
                    [self setPerson: [[chain lastObject] intValue] to: SPY];
                    //[identities replaceObjectAtIndex: [[chain lastObject] intValue]
                    //                      withObject: [NSNumber numberWithInt: SPY]];
                    //NSLog(@"setting last to spy");
                }
                else {
                    [self setPerson: [[chain lastObject] intValue] to: KNIGHT];

                    //[identities replaceObjectAtIndex: [[chain lastObject] intValue]
                    //                      withObject: [NSNumber numberWithInt: KNIGHT]];
                    //NSLog(@"setting last to knight");
                }
            }
        }
		//NSLog(@"identities = %@",identities);
    }
    //no need to return anything: identities modified
}

-(int) determineFirstKnightInTaggedChain: (TaggedChain*) taggedChain
                  persistWithShortChains: (BOOL) persist
{
    //NSLog(@"determineFirstKnightInTaggedChain called with taggedChain = %@",taggedChain);
    NSMutableArray* chainCopy = [[taggedChain chain] mutableCopy];
    int firstKnight = [chainCopy count] - 1;
    BOOL foundKnight;

    //if ([taggedChain endsWithSpy]) {
    //    NSLog(@"dealing with a chain ending in a claimed spy");
    [chainCopy removeLastObject]; //its impossible we know anything about this chap (unless he's a K)
    //its possible that we know something about the people in what remains (we could know about
    //more people than just the last remaining --- BUG)
    if ([chainCopy count] == 0) {
        if (!persist) return -1; else return 0;
        //we return 0, which will lead to an identity being asked only if we are in persistent mode
    }
    
    int lastNumber = [[chainCopy lastObject] intValue];
    //NSLog(@"number in penultimate position = %i",number);
    if ([[identities objectAtIndex: lastNumber] intValue] == SPY) {
        //NSLog(@"Penultimate member known to be a spy");
        //we return firstKnight as it stands (this doesn't necessarily mean the last person is a Knight)
        //NSLog(@"returning firstKnight = %i",firstKnight);
        return firstKnight;
    }
    //if here then the chain definitely doesn't end with a spy
    int i = 0; while ([[identities objectAtIndex: [[chainCopy objectAtIndex: 0] intValue]] intValue] == SPY) {
        [chainCopy removeObjectAtIndex: 0];
        i+=1;
    }
    while (([[identities objectAtIndex: [[chainCopy lastObject] intValue]] intValue] == KNIGHT)
           && ([chainCopy count] != 0)) {
        //   && (firstKnight >= 1)) {
        //vicious bug that arrays are terminated by 0. So we have to take care not to
        //step past the `end' of the array
        //the initial correction is no good, as the chain could be empty at this point
        //NSLog(@"firstKnight = %i",firstKnight);
        //NSLog(@"Removing known knight %i",[[chainCopy lastObject] intValue]);
        [chainCopy removeLastObject];
        foundKnight = TRUE;
        firstKnight -=1;
        //NSLog(@"Now firstKnight = %i",firstKnight);
    }
    

    //NSLog(@"After removing any initial spies and final knights, chain = %@",chainCopy);
    
    if (([chainCopy count] <= 3) && !persist) { // && !foundKnight) { SHORT CHAIN DEFINITION
                                    //in runs on 8/7 needed 3 to avoid 23 Qs but this gave a 31 on 11/9
                                    //NSLog(@"This is a short chain, not persisting");
                                    //if have found a knight, want to carry on as can fill in final entry at least
                                    //still this may not be the most efficient approach
        return -1;
    }
    else
    {
        int firstKnightInChainCopy = [self determineFirstKnightIn: chainCopy];
        if (firstKnightInChainCopy == [chainCopy count]) {
            //NSLog(@"no knights in chainCopy");
            //no need to modify firstKnight
        }
        else {
            //NSLog(@"first knight in chainCopy at %i",firstKnightInChainCopy);
            firstKnight = firstKnightInChainCopy+i;
        }
        //NSLog(@"returning firstKnight = %i",firstKnight);
        return firstKnight; //had wrongly been firstKnightInChainCopy
    }
}


-(int) determineFirstKnightIn: (NSMutableArray*) chain
{
    //NSLog(@"determineFirstKnightIn ... called with chain = %@",chain);
    int n = [chain count]; int m;
    if (n==0) {
        //NSLog(@"the chain is empty");
        return 0;
    }
    if (n==1) {
        //NSLog(@"left with only one person, namely %i",[[chain objectAtIndex: 0] intValue]);
        BOOL isKnight = ([[knight viewOnPerson: [[chain objectAtIndex: 0] intValue]] intValue] == KNIGHT);
        //NSLog(@"BOOL isKnight set");
        if (isKnight)
            return 0;
        else
            return 1;
    }
    else {
		if (n==2) m=0; else m=(n-1)/2; //n=2 arises frequently with KKS chains, and we want to ask in middle? but worse in general?
        //m=(n-1)/2; 
		int i; //so e.g. n=5, m=2, XXYXX, n=6, m=2 XXYXXX
                          //NSLog(@"going to ask about person %i",[[chain objectAtIndex: m] intValue]);
        NSMutableArray* newChain = [chain mutableCopy];
        if ([[knight viewOnPerson: [[chain objectAtIndex: m] intValue]] intValue] == SPY) {
            //NSLog(@"He was a spy");
            for (i=0;i<=m;i++) //m+1 people removed, namely 0, 1, ... m
                [newChain removeObjectAtIndex: 0];
            return m + 1 + [self determineFirstKnightIn: newChain];
        }
        else {
            //NSLog(@"He was a knight");
            for (i=m;i<n;i++) // n-m people removed, namely m, m+1, ... n -1
                [newChain removeObjectAtIndex: m];
            return [self determineFirstKnightIn: newChain];
        }
    }
}


-(void) setPerson: (int) p to: (int) IDENT
    //this function should be called to set all identities. It pursues `knock on consequences' in chains
    //only use it to set IDENT to KNIGHT / SPY, not UNKNOWN
{
    [identities replaceObjectAtIndex: p
                          withObject: [NSNumber numberWithInt: IDENT]];
    //NSLog(@"Just made person %i a %@",p,(IDENT == KNIGHT) ? @"Knight": @"Spy");
    //now iterate through chains, looking through chains containing person i
    //doesn't matter which order we go in, i.e. we don't attempt to puruse `second order'
    //consequences. These would be picked up later anyway.
    int i; for (i=0;i<[allChains count];i++) {
        NSMutableArray* chains = [allChains objectAtIndex: i];
        int j; for (j=0;j<[chains count];j++) {
            TaggedChain* taggedChain = [chains objectAtIndex: j];
            NSMutableArray* chain = [taggedChain chain];
            //NSLog(@"Looking for person %i in  chain %@",p,chain);
            BOOL ourChapInChain = FALSE; int position;
            int k; for (k=0;k<[chain count];k++) {
                if ([[chain objectAtIndex: k] intValue] == p) {
                    ourChapInChain = TRUE; position = k;
                }
            }
            if (ourChapInChain) {
                //NSLog(@"Chain %@ contains our person in position %i",chain,position);
                if (IDENT == KNIGHT) {
                    //NSLog(@"Our person is a Knight, so setting going forward in chain");
                    //NSLog(@"position = %i,[chain count] =%i",position,[chain count]);
                    int l; for (l=position+1;l<[chain count]-1;l++) {
                        [identities replaceObjectAtIndex: [[chain objectAtIndex: l] intValue]
                                              withObject: [NSNumber numberWithInt: KNIGHT]];
                        //NSLog(@"Just made person %i a knight",[[chain objectAtIndex: l] intValue]);
                    }
                    //NSLog(@"Now dealing with last chap");
                    int numberOfLastChap = [[chain lastObject] intValue];
                    //NSLog(@"Number of last chap = %i",numberOfLastChap);
                    //if called with our Knight at the end, don't want to do anything
                    //this really is absolutely dreadful
                    if (position < [chain count] - 1) {
                        if ([taggedChain endsWithSpy]) {
                            [identities replaceObjectAtIndex: numberOfLastChap
                                                  withObject: [NSNumber numberWithInt: SPY]];
                            //NSLog(@"Just made person %i, last in chain, a spy",numberOfLastChap);
                        }
                        else {
                            [identities replaceObjectAtIndex: numberOfLastChap
                                                  withObject: [NSNumber numberWithInt: KNIGHT]];
                            //NSLog(@"Just made person %i, last in chain, a knight",numberOfLastChap);
                        }
                    }
                    //NSLog(@"Dealt with last chap");
                }
                else {
                    //if we've been given a spy at end chain, don't want to do anything
                    if (position < [chain count]-1) {
                        int l; for (l=0;l<position;l++) {
                            [identities replaceObjectAtIndex: [[chain objectAtIndex: l] intValue]
                                                  withObject: [NSNumber numberWithInt: SPY]];
                            //NSLog(@"Just made person %i a spy",[[chain objectAtIndex: l] intValue]);
                        }
                    }
                }
            }
        }
    }
}


@end
