//  Crowd.m
//  KnightsAndSpies

#import "Crowd.h"


@implementation Crowd
-(Person *) person: (int) i {
    return [members objectAtIndex: i]; }

-(int) numberOfPeople {
    return [members count]; }

-(Crowd *) initWithGivenMembers: (NSMutableArray*) givenMembers {
    self = [super init];
    numberOfPeople = [givenMembers count];
    members = [[NSMutableArray alloc] init];
    int i; for (i=0;i<numberOfPeople;i++) {
        [members addObject:
            ([[givenMembers objectAtIndex: i] intValue] == KNIGHT) ?
            [[Knight alloc] initWithNumber: i] :
            [[Spy alloc] initWithNumber: i]];
    }
    return self;
}

-(Crowd *) initRandomlyWithKnights: (int) knights spies: (int) spies {
    self = [super init];
    numberOfPeople = knights + spies;
    members = [[NSMutableArray alloc] init];

    int i; for (i=0;i<numberOfPeople;i++) {
        [members addObject: [[Knight alloc] initWithNumber: i]];
    }
    //NSLog(@"Allocated the following crowd");
    //[self showMembers];

    //NSLog(@"Now replacing with spies",numberOfPeople);
    int spiesAllocated = 0;
    while (spiesAllocated < spies) {
        BOOL newSpyAllocated = FALSE;
        while (newSpyAllocated == FALSE) {
            int newSpy = rand() % numberOfPeople;
            //NSLog(@"Attempting to make %i a spy",newSpy);
            if ([[self person: newSpy] isKnight]) {
                [members replaceObjectAtIndex: newSpy withObject:
                    [[Spy alloc] initWithNumber: newSpy]];
                newSpyAllocated = TRUE;
                //NSLog(@"Succeeded!");
            }
        }
        spiesAllocated++;
    }
    
    return self;
}

-(void) setKnightsViews {
    NSMutableArray* trueSituation = [self membersAsNumbers];
    int i; for (i=0;i<numberOfPeople;i++) {
        Person* candidate = [members objectAtIndex : i];
        if ([candidate isKnight]) {
            [candidate setViews: trueSituation];
        }
    }
}

-(void) setSpiesViewsToTruth {
    NSMutableArray* trueSituation = [self membersAsNumbers];
    int i; for (i=0;i<numberOfPeople;i++) {
        Person* candidate = [members objectAtIndex : i];
        if ([candidate isSpy]) {
            [candidate setViews: trueSituation];
        }
    }
}

-(void) setSpiesViewsToAlwaysLie {
    NSMutableArray* trueSituation = [self membersAsNumbers];
    int i; for (i=0;i<numberOfPeople;i++) {
        Person* candidate = [members objectAtIndex : i];
        if ([candidate isSpy]) {
            [candidate setViewsToAlwaysLie: trueSituation];
        }
    }
}

-(void) setSpiesViewsRandomly {
    int i; for (i=0;i<numberOfPeople;i++) {
        Person* candidate = [members objectAtIndex : i];
        if ([candidate isSpy]) {
            [candidate setViewsRandomlyWithCrowdSize: numberOfPeople];
        }
    }
}

-(void) setSpiesViewsToAllKnights {
    //NSLog(@"setSpiesViewsToAllKnights here");
    int i; for (i=0;i<numberOfPeople;i++) {
        Person* candidate = [members objectAtIndex : i];
        if ([candidate isSpy]) {
            //NSLog(@"sSVTAK setting views of spy %i",i);
            [candidate setViewsToAllKnightsWithCrowdSize: numberOfPeople];
        }
    }
}

-(void) setSpiesViewsToAllSpies {
    int i; for (i=0;i<numberOfPeople;i++) {
        Person* candidate = [members objectAtIndex : i];
        if ([candidate isSpy]) {
            [candidate setViewsToAllSpiesWithCrowdSize: numberOfPeople];
        }
    }
}

-(void) setSpiesViewsToReverseOrSpy {
	//NSLog(@"setSpiesViewsToReverseOrSpy here");
    NSMutableArray* reverseSituation = [self membersAsNumbers];
    int i; for (i=0;i<numberOfPeople;i++) {
        [reverseSituation replaceObjectAtIndex: i
                                    withObject: ([[reverseSituation objectAtIndex: i] intValue]) == KNIGHT
              ? [NSNumber numberWithInt: SPY] : [NSNumber numberWithInt: KNIGHT]];
    }
    for (i=0;i<numberOfPeople;i++) {
        Person* candidate = [members objectAtIndex : i];
        if ([candidate isSpy]) {
			if (rand() % 2 == 1) {
				//NSLog(@"reverse");
				[candidate setViews: reverseSituation];
			}
			else {
				//NSLog(@"spy");
				[candidate setViewsToAllSpiesWithCrowdSize: numberOfPeople];
			}
        }
    }
}

-(void) setSpiesViewsReversely {
    NSMutableArray* reverseSituation = [self membersAsNumbers];
    int i; for (i=0;i<numberOfPeople;i++) {
        [reverseSituation replaceObjectAtIndex: i
                                    withObject: ([[reverseSituation objectAtIndex: i] intValue]) == KNIGHT
              ? [NSNumber numberWithInt: SPY] : [NSNumber numberWithInt: KNIGHT]];
    }
    for (i=0;i<numberOfPeople;i++) {
        Person* candidate = [members objectAtIndex : i];
        if ([candidate isSpy]) {
            [candidate setViews: reverseSituation];
        }
    }
}

-(void) showMembers {
    NSLog(@"members = %@",members); }

-(void) showMembersFully {
    int i; for (i=0;i<[members count]; i++) {
        Person* candidate = [members objectAtIndex: i];
        ([candidate isKnight]) ? NSLog(@"Person %i is a knight",i) :
            NSLog(@"Person %i is a spy",i);
        [candidate showViews];
        [candidate showQuestionsAsked];
    }
}

-(void) showMembersAbbreviatedly {
    [self showMembers];
    int i; for (i=0;i<[members count]; i++) {
        Person* candidate = [members objectAtIndex: i];
        ([candidate isKnight]) ? NSLog(@"Person %i is a knight",i) :
            NSLog(@"Person %i is a spy",i);
        [candidate showViewsAbbreviatedly];
    }
}


-(int) numberOfQuestionsAsked
{
    int numberOfQuestionsAsked = 0;
    int i; for (i=0;i<[self numberOfPeople];i++) {
        numberOfQuestionsAsked += [[self person: i] numberOfQuestionsAsked];
    }
    return numberOfQuestionsAsked;
}

//Use this for setting views (views are arrays of NSNumbers)
-(NSMutableArray *) membersAsNumbers {
    NSMutableArray* people = [[NSMutableArray alloc] init];
    int i; for (i=0;i<[members count];i++) {
        [people addObject:
            ([[members objectAtIndex: i] isKnight]) ?
             [NSNumber numberWithInt: KNIGHT] :
             [NSNumber numberWithInt: SPY]];
    }
    return people;
}

-(BOOL) testIfCorrectAnswer: (NSMutableArray*) answer
{
    BOOL isCorrect = TRUE;
    int i; for (i=0;i<[members count];i++) {
        BOOL knightWrong = ([[members objectAtIndex: i] isKnight])
        && ([[answer objectAtIndex: i] intValue] != KNIGHT);
        BOOL spyWrong = ([[members objectAtIndex: i] isSpy])
            && ([[answer objectAtIndex: i] intValue] != SPY);
        if (knightWrong || spyWrong) {
            NSLog(@"Given answer is wrong at person %i",i);
            isCorrect = FALSE;
        }
    }
    return isCorrect;
}

-(BOOL) someoneAskedTwiceAboutSamePerson
{
    int i; BOOL askedTwice = FALSE;
    for (i=0;i<[members count];i++) {
        Person* person = [self person: i];
        BOOL thisPersonAskedTwice = [person askedTwiceAboutSamePerson];
        if (thisPersonAskedTwice) {
            //NSLog(@"Person %i asked twice, %@",[person number],[person questionsAsked]);
        }
        askedTwice = askedTwice || thisPersonAskedTwice;
    }
    return askedTwice;
}

-(void) reset
{
    //NSLog(@"reset crowd called");
    int i; for (i=0;i<[members count];i++) {
        //NSLog(@"i = %i",i);
        Person* person = [self person: i];
        [person resetQuestionsAsked];
    }
}
        


@end
