//  Interrogator.m
//  KnightsAndSpies

#import "Interrogator.h"
#define DEBUG

extern void NSLogD (NSString *format, ...);

extern int DWCB;
extern int MWCB;
extern int NAIVECR;
extern int ALLCR;
extern int LONGCR;
extern int JOINCR;

@implementation Interrogator

-(Interrogator*) initWithCrowd: (Crowd*) givenCrowd
{
    self = [super init];
    crowd = givenCrowd; //Not copied
    //NSLog(@"initWithCrowd here, just set weights = %@",weights);
    //NSLog(@"crowd = %@",crowd);[crowd showMembers];
    return self;
}

-(void) setChainBuilder: (int) givencb setChainResolver: (int) givencr
{
    cb = givencb;
    cr = givencr;
}

-(NSDictionary*) determineIdentitiesAcceptingAt: (int) l
{
    NSDictionary* results;
    if (cb == DWCB) {
        ChainBuilderDW* chainBuilder = [[ChainBuilderDW alloc] initWithDW: crowd];
        results = [chainBuilder buildChains]; //knightchains, spyChains, joinedChains, finalKnight
    }
    else if (cb == MWCB) {
        ChainBuilderMW* chainBuilder = [[ChainBuilderMW alloc] initWithMW: crowd acceptingAt: l];
        results = [chainBuilder buildChains]; //allChains, joinedChains, finalKnight
    }
    else {
        NSLog(@"UNKNOWN CHAINBUILDER REQUESTED");
    }

    //extract results
    NSMutableArray *allChains = [results objectForKey: @"allChains"];
    int finalKnight = [[results objectForKey: @"finalKnight"] intValue];
    Person *knight = [crowd person: finalKnight];

    int questionsToKnight = [crowd numberOfQuestionsAsked];
    
    //ckeck that disaster hasn't occurred already
    if ([knight isSpy]) {
        NSLog(@"FINISHED AT SPY %i after %i questions",finalKnight,[crowd numberOfQuestionsAsked]);
        NSLog(@"ENDED AT SPY. Crowd following.");
        [crowd showMembersAbbreviatedly];
        NSLog(@"allChains = %@", allChains);
        NSLog(@"");
    }
    else {
        //comment out
        //NSLog(@"Finished at person %i after %i questions",finalKnight,[crowd numberOfQuestionsAsked]);
        //NSLog(@"allChains = %@, but we might join them later ...",allChains);
        //NSLog(@"Used %i questions to get to knight",[crowd numberOfQuestionsAsked]);
    }

    ChainResolver* chainResolver = [[ChainResolver alloc] init];
    NSMutableArray* identities;
    if (cr == NAIVECR) {
        //This method is really too naive. For instance, suppose we have K S, K
        //then the chains are ((0,1), (2)), (0,2) (although the last is actually redundant)
        //and will ask again about 2, as we never look at the (0,2) chain
        [chainResolver initWith: crowd allChains: allChains knight: knight];
        NSMutableArray* firstChains = [allChains objectAtIndex: 0];
        //NSLog(@"firstChains = %@",firstChains);
        [chainResolver addToIdentitiesUsingChains: firstChains resolvingEnds: TRUE];
        identities = [chainResolver identities];

    }
    else if (cr == JOINCR) { //only one we are still working on
        ChainJoiner* chainJoiner = [[ChainJoiner alloc] initWithAllChains: allChains];
        NSMutableArray* joinedChains = [chainJoiner joinChains];
        //NSLog(@"JOINCR, joinedChains = %@",joinedChains);
		
        [chainResolver initWith: crowd allChains: joinedChains knight: knight];
        identities = [chainResolver resolveChains];
		//NSLog(@"identities = %@",identities);
		//it is possible that some identities will still be ambiguous if we accepted a knight in the linear chain building phase
		//we simply look through identities for -1s and question our knight about them
		int i; for (i=0;i<[identities count];i++) {
			if ([[identities objectAtIndex: i] intValue] == -1) {
				NSNumber* view = [knight viewOnPerson: i];
				[identities replaceObjectAtIndex: i withObject: view];
				//NSLog(@"Identity of person %i was unknown, now known to be %@",i,view);
			}
		}
		//now should be finished
    }
    else { //in all other cases, it makes sense to join up chains iff the method is DW
        //so ALL in the context of DW means resolve ALL after joining
        //LONGJOIN means resolve LONG after joining
        //this could be clearer
        if (cb == DWCB) {
            ChainJoiner* chainJoiner = [[ChainJoiner alloc] initWithAllChains: allChains];
            NSMutableArray* joinedChains = [chainJoiner joinChains];
            allChains = joinedChains;
        }
        //NSLog(@"cb = %i, cr = %i",cb,cr);
        //NSLog(@"allChains = %@",allChains);

        if (cr == ALLCR) {
        [chainResolver initWith: crowd allChains: allChains knight: knight];
        identities = [chainResolver resolveChainsAll];
        }
        else if (cr == LONGCR) {
            [chainResolver initWith: crowd allChains: allChains knight: knight];
            identities = [chainResolver resolveChains];
        }
        else
            NSLog(@"Requested ChainResolver not yet coded in interrogator.m");
    }
        
    return [NSDictionary dictionaryWithObjectsAndKeys: identities, @"identities",
           [NSNumber numberWithInt: questionsToKnight], @"questionsToKnight", nil];
        
    
    //ChainResolver *chainResolverJoinedChains =
    //    [[ChainResolver alloc] initWith: crowd allChains: joinedChains knight: knight];
    //NSMutableArray *identities = [chainResolverJoinedChains resolveChains];
    //NSLog(@"Used %i questions joining chains",[crowd numberOfQuestionsAsked]);

    
    //if ([crowd numberOfQuestionsAsked] == 22) {
    //    NSLog(@"22 questions");
    //    [crowd showMembersAbbreviatedly];
    //    NSLog(@"allChains =%@",allChains);
    //}
    //return identities;
    
}

@end
