//  Tester.m
//  KnightsAndSpies


#import "Tester.h"

//need all of these for multipleTests method
extern int DWCB;
extern int MWCB;
extern int NAIVECR;
extern int ALLCR;
extern int LONGCR;
extern int JOINCR;

@implementation Tester

-(NSDictionary*) singleTestWithCrowd: (Crowd*) crowd spiesViews: (int) spiesViews
                          chainBuilder: (int) cb chainResolver: (int) cr acceptingAt: (int) l
{
    [crowd setKnightsViews];
    if (spiesViews == KNIGHT) {
        [crowd setSpiesViewsToAllKnights];
    }
    else if (spiesViews == SPY) {
        [crowd setSpiesViewsToAllSpies];
    }
    else if (spiesViews == RANDOM) {
        [crowd setSpiesViewsRandomly];
    }
    else if (spiesViews == REVERSE) {
        [crowd setSpiesViewsReversely];
    }
    else if (spiesViews == TRUTH) {
        [crowd setSpiesViewsToTruth];
    }    
    else if (spiesViews == REVERSEORSPY) {
		//NSLog(@"About to initialise spies views to REVERSEORSPY");
        [crowd setSpiesViewsToReverseOrSpy];
    }    
    
	
    Interrogator* interrogator = [[Interrogator alloc] initWithCrowd: crowd];
    [interrogator setChainBuilder: cb setChainResolver: cr];

    NSDictionary* outcome = [interrogator determineIdentitiesAcceptingAt: l];
    NSMutableArray* identities = [outcome objectForKey: @"identities"];
    int questionsToKnight = [[outcome objectForKey: @"questionsToKnight"] intValue];

    BOOL gotItRight = [crowd testIfCorrectAnswer: identities];

    if (gotItRight) {
        //NSLog(@"Success. With %i then %i questions. Total %i",
          //    questionsToKnight,[crowd numberOfQuestionsAsked]-questionsToKnight //nb questionsToKnight means questions to reach the knight
          //    [crowd numberOfQuestionsAsked]);
        //NSLog(@"Success. Number of questions was: %i",[crowd numberOfQuestionsAsked]);
        //[crowd showMembersFully];
    }
    else
    {
        NSLog(@"INCORRECT ANSWER; cb = %i, cr = %i",cb,cr);
        [crowd showMembersAbbreviatedly]; //don't report allViews, only those that had
                                          //some bearing on the dreadful outcome
        NSLog(@"identities = %@",identities);
    }
    if ([crowd numberOfQuestionsAsked] > [crowd numberOfPeople] + l - 1) { //3*[crowd numberOfPeople]/2) {
        NSLog(@"TOO MANY QUESTIONS, namely %i; cb = %i, cr = %i",[crowd numberOfQuestionsAsked],cb,cr);
        [crowd showMembersAbbreviatedly];
    }
    if ([crowd someoneAskedTwiceAboutSamePerson] && (cr != NAIVECR)) {
        //this is expected if using NAIVECR
        NSLog(@"SOMEONE ASKED TWICE ABOUT SAME PERSON");
        [crowd showMembersAbbreviatedly];
    }

    NSDictionary* results = [NSDictionary dictionaryWithObjectsAndKeys:
        [NSNumber numberWithInt: [crowd numberOfQuestionsAsked]], @"numberOfQuestions",
        [NSNumber numberWithInt: questionsToKnight], @"questionsToKnight",
        [NSNumber numberWithBool: gotItRight], @"gotItRight", nil];
    //[crowd release];
    return results;
}

-(NSDictionary*) spiderTestWithCrowd: (Crowd*) crowd spiesViews: (int) spiesViews acceptingAt: (int) l
{
    //NSLog(@"spiderTestWithCrowd here");
    //code here is a subset of the code in singleTest
    [crowd setKnightsViews];
    if (spiesViews == KNIGHT) {
        [crowd setSpiesViewsToAllKnights];
    }
    else if (spiesViews == SPY) {
        [crowd setSpiesViewsToAllSpies];
    }
    else if (spiesViews == RANDOM) {
        [crowd setSpiesViewsRandomly];
    }
    else if (spiesViews == REVERSE) {
        [crowd setSpiesViewsReversely];
    }
    else if (spiesViews == TRUTH) {
        [crowd setSpiesViewsToTruth];
    }     //all the same as singleTest
	else if (spiesViews == REVERSEORSPY) {
		//NSLog(@"About to initialise spies views to REVERSEORSPY");
        [crowd setSpiesViewsToReverseOrSpy];
    }       

    
    SpiderInterrogator* interrogator = [[SpiderInterrogator alloc] initWithCrowdSpider: crowd];
    NSDictionary* outcome = [interrogator determineIdentitiesAcceptingAt: l];
    NSMutableArray* identities = [outcome objectForKey: @"identities"];
    int questionsToKnight = [[outcome objectForKey: @"questionsToKnight"] intValue];
    int questionsSaved = [[outcome objectForKey: @"questionsSaved"] intValue];

    BOOL gotItRight = [crowd testIfCorrectAnswer: identities];

    if (gotItRight) {
        //NSLog(@"Success. With %i then %i questions. Total %i. Questions saved %i",
	    //   [crowd numberOfQuestionsAsked]-questionsToKnight,questionsToKnight,
		//   [crowd numberOfQuestionsAsked],questionsSaved);
        //NSLog(@"Success. Number of questions was: %i",[crowd numberOfQuestionsAsked]);
        //[crowd showMembersAbbreviatedly];
    }
    else
    {
        NSLog(@"INCORRECT ANSWER");
        [crowd showMembersAbbreviatedly]; //don't report allViews, only those that had
                                          //some bearing on the dreadful outcome
        NSLog(@"identities = %@",identities);
    }
    if ([crowd numberOfQuestionsAsked] > [crowd numberOfPeople] + l - 1) {
        NSLog(@"TOO MANY QUESTIONS, namely %i",[crowd numberOfQuestionsAsked]);
        [crowd showMembersAbbreviatedly];
    }
    if ([crowd someoneAskedTwiceAboutSamePerson]) {
        NSLog(@"SOMEONE ASKED TWICE ABOUT SAME PERSON");
        [crowd showMembersAbbreviatedly];
    }

    NSDictionary* results = [NSDictionary dictionaryWithObjectsAndKeys:
        [NSNumber numberWithInt: [crowd numberOfQuestionsAsked]], @"numberOfQuestions",
        [NSNumber numberWithInt: questionsToKnight], @"questionsToKnight",
        [NSNumber numberWithBool: gotItRight], @"gotItRight", nil];
    //[crowd release];
    return results;

}

-(NSDictionary*) singleTestWithKnights: (int) knights spies: (int) spies spiesViews: (int) spiesViews
                                 chainBuilder: (int) cb chainResolver: (int) cr acceptingAt: (int) l
{
    Crowd* crowd = [[Crowd alloc] initRandomlyWithKnights: knights spies: spies];
    return [self singleTestWithCrowd: crowd spiesViews: spiesViews
                 chainBuilder: cb chainResolver: cr acceptingAt: l];
}

-(NSDictionary*) spiderTestWithKnights: (int) knights spies: (int) spies
                            spiesViews: (int) spiesViews acceptingAt: (int) l
{
    Crowd* crowd = [[Crowd alloc] initRandomlyWithKnights: knights spies: spies];
    return [self spiderTestWithCrowd: crowd spiesViews: spiesViews acceptingAt: l];
}

-(void) multipleTestsWithKnights: (int) knights spies: (int) spies spiesViews: (int) spiesViews N: (int) N
                    chainBuilder: (int) cb chainResolver: (int) cr acceptingAt: (int) l
{
    //make a table to store the results
    NSMutableArray *table = [[NSMutableArray alloc] init];
    int j; for (j=0;j<2*(knights+spies);j++){
        [table addObject: [NSNumber numberWithInt: 0]];
    }

    int i; for (i=0;i<N;i++) {
        NSDictionary* results = [self singleTestWithKnights: knights spies: spies spiesViews: spiesViews
                                               chainBuilder: cb chainResolver: cr acceptingAt: l];
        int questionsAsked = [[results objectForKey: @"numberOfQuestions"] intValue];
        //int questionsToKnight = [[results objectForKey: @"questionsToKnight"] intValue];
        BOOL gotItRight = [[results objectForKey: @"gotItRight"] boolValue];
        if (!gotItRight) {
            NSLog(@"Stopping tests");
            break;
        }
        else {
            int r = [[table objectAtIndex: questionsAsked] intValue];
            [table replaceObjectAtIndex: questionsAsked
                             withObject: [NSNumber numberWithInt: r+1]];
            if ((i > 0) && (i % 500 == 0)) { NSLog(@"Iteration: %i",i); }
        }
    }

    NSLog(@"");
    BOOL actionStarted = FALSE; int acc = 0; float tot = 0;
    for (i=0;i<2*(knights+spies);i++) {
        int n = [[table objectAtIndex: i] intValue];
        acc += n; tot += i * n;
        if (n > 0)
            actionStarted = TRUE;
        if (actionStarted)
            NSLog(@"%i, %i",i,n);
        if (acc == N)
            actionStarted = FALSE;
    }
    NSLog(@"Mean number of questions = %f",tot/N);
}

//haven't updated since we never use it these days
-(void) multipleTestsWithKnights: (int) knights spies: (int) spies spiesViews: (int) spiesViews N: (int) N
{
    NSLog(@"DWALL, DWLONGJOIN, MWALL, MWLONG, MWJOIN");
    int i; for (i=0;i<N;i++) {
        Crowd* crowd = [[Crowd alloc] initRandomlyWithKnights: knights spies: spies];
        NSDictionary* results = [self singleTestWithCrowd: crowd spiesViews: spiesViews
                                               chainBuilder: DWCB chainResolver: ALLCR];
        int qDWALL = [[results objectForKey: @"numberOfQuestions"] intValue];
        BOOL gotItRight = [[results objectForKey: @"gotItRight"] boolValue];
        [crowd reset];

        results = [self singleTestWithCrowd: crowd spiesViews: spiesViews                                               		chainBuilder: DWCB chainResolver: JOINCR];
        int qDWLONG = [[results objectForKey: @"numberOfQuestions"] intValue];
        gotItRight = [[results objectForKey: @"gotItRight"] boolValue] && gotItRight;
        [crowd reset];

        //don't bother with DWCB LONGCR because same as last case (if there
        //are to be long chains we have to join them up)
        //if don't join up then there are no LONG chains so becomes same
        //as naive method

        results = [self singleTestWithCrowd: crowd spiesViews: spiesViews	                                               		chainBuilder: MWCB chainResolver: ALLCR];
        int qMWALL = [[results objectForKey: @"numberOfQuestions"] intValue];
        gotItRight = [[results objectForKey: @"gotItRight"] boolValue] && gotItRight;
        [crowd reset];

        results = [self singleTestWithCrowd: crowd spiesViews: spiesViews                                               		chainBuilder: MWCB chainResolver: LONGCR];
        int qMWLONG = [[results objectForKey: @"numberOfQuestions"] intValue];
        gotItRight = [[results objectForKey: @"gotItRight"] boolValue] && gotItRight;
        [crowd reset];

        results = [self singleTestWithCrowd: crowd spiesViews: spiesViews                                               		chainBuilder: MWCB chainResolver: JOINCR];
        int qMWJOIN = [[results objectForKey: @"numberOfQuestions"] intValue];
        gotItRight = [[results objectForKey: @"gotItRight"] boolValue] && gotItRight;
        [crowd reset];
        
        //int questionsToKnight = [[results objectForKey: @"questionsToKnight"] intValue];

        if (!gotItRight) {
            NSLog(@"Stopping tests, someone got it wrong");
            break;
        }
        else {
            NSLog(@"%i, %i, %i, %i, %i",qDWALL, qDWLONG, qMWALL, qMWLONG, qMWJOIN); 
        }
    }
}

-(void) multipleSpiderTestsFindingKnightWithKnights: (int) knights spies: (int) spies
										spiesViews: (int) spiesViews N: (int) N acceptingAt: (int) l
{
	NSMutableArray *table = [[NSMutableArray alloc] init];
	int j; for (j=0;j<knights+spies-1;j++) { //we know that at most 2*max spies - 1 questions suffice
		[table addObject: [NSNumber numberWithInt: 0]];
	}
	
	int i; for (i=0;i<N;i++) {
		NSDictionary* results = [self spiderTestWithKnights: knights spies: spies spiesViews: spiesViews
												acceptingAt: l];
        int questionsToFindKnight = [[results objectForKey: @"numberOfQuestions"] intValue] -
			[[results objectForKey: @"questionsToKnight"] intValue];
		BOOL gotItRight = [[results objectForKey: @"gotItRight"] boolValue];
		if (!gotItRight) {
			NSLog(@"Stopping tests");
			break;
		}
		else {
			int r = [[table objectAtIndex: questionsToFindKnight] intValue];
			[table replaceObjectAtIndex: questionsToFindKnight withObject: [NSNumber numberWithInt: r+1]];
			if ((i > 0) && (i % 500 == 0)) { NSLog(@"Iteration: %i",i); }
		}
	}
	
	NSLog(@"");
    BOOL actionStarted = FALSE; int acc = 0; float tot = 0;
    for (i=0;i<knights+spies-1;i++) {
        int n = [[table objectAtIndex: i] intValue];
        acc += n; tot += i * n;
        if (n > 0)
            actionStarted = TRUE;
        if (actionStarted)
            NSLog(@"%i, %i",i,n);
        if (acc == N)
            actionStarted = FALSE;
    }
    NSLog(@"Mean number of questions = %f",tot/N);
}	
							

-(void) multipleSpiderTestsWithKnights: (int) knights spies: (int) spies
                            spiesViews: (int) spiesViews N: (int) N acceptingAt: (int) l
{
    //make a table to store the results
    NSMutableArray *table = [[NSMutableArray alloc] init];
    int j; for (j=0;j<2*(knights+spies);j++){
        [table addObject: [NSNumber numberWithInt: 0]];
    }

    int i; for (i=0;i<N;i++) {
        NSDictionary* results = [self spiderTestWithKnights: knights spies: spies spiesViews: spiesViews
                                                acceptingAt: l];
        int questionsAsked = [[results objectForKey: @"numberOfQuestions"] intValue];
        //int questionsToKnight = [[results objectForKey: @"questionsToKnight"] intValue];
        BOOL gotItRight = [[results objectForKey: @"gotItRight"] boolValue];
        if (!gotItRight) {
            NSLog(@"Stopping tests");
            break;
        }
        else {
            int r = [[table objectAtIndex: questionsAsked] intValue];
            [table replaceObjectAtIndex: questionsAsked
                             withObject: [NSNumber numberWithInt: r+1]];
            if ((i > 0) && (i % 500 == 0)) { NSLog(@"Iteration: %i",i); }
        }
    }

    NSLog(@"");
    BOOL actionStarted = FALSE; int acc = 0; float tot = 0;
    for (i=0;i<2*(knights+spies);i++) {
        int n = [[table objectAtIndex: i] intValue];
        acc += n; tot += i * n;
        if (n > 0)
            actionStarted = TRUE;
        if (actionStarted)
            NSLog(@"%i, %i",i,n);
        if (acc == N)
            actionStarted = FALSE;
    }
    NSLog(@"Mean number of questions = %f",tot/N);
}


-(void) specialTest
{
    NSMutableArray* testMembers = [[NSMutableArray alloc] init];
    [testMembers addObject: [NSNumber numberWithInt: SPY]];
    [testMembers addObject: [NSNumber numberWithInt: SPY]];
    [testMembers addObject: [NSNumber numberWithInt: SPY]];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: SPY]];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: SPY]];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: SPY]];
    Crowd* testCrowd = [[Crowd alloc] initWithGivenMembers: testMembers];
    [testCrowd setSpiesViewsReversely]; [testCrowd setKnightsViews];
    NSLog(@"testCrowd = %@",testCrowd); [testCrowd showMembers];

    //won't work at moment
}

-(void) crosslinkTest {
   NSMutableArray* testMembers = [[NSMutableArray alloc] init];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //1
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //2
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //3
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //4
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //5
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //6
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //7 //can accept at this piont, but a chain ending at 10 will be built
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //8
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //9
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //10	
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //11	
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //12		
	Crowd* testCrowd = [[Crowd alloc] initWithGivenMembers: testMembers];
	NSDictionary* results = [self singleTestWithCrowd: testCrowd spiesViews: REVERSE
								  chainBuilder: MWCB chainResolver: JOINCR acceptingAt: 5];
	int qs = [[results objectForKey: @"numberOfQuestions"] intValue];
    BOOL gotItRight = [[results objectForKey: @"gotItRight"] boolValue];
	NSLog(@"Number of questions = %i, gotItRight = %i",qs,gotItRight);
	[testCrowd showMembersFully];
}

-(void) smallTest
{
    NSMutableArray* testMembers = [[NSMutableArray alloc] init];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //1
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //2
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //3
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //4
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //5
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //6
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //7 //can and will accept at this point. note only linear chains, and not everyone is in a chain
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //8
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //9
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //10	
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //11	
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //12		
	Crowd* testCrowd = [[Crowd alloc] initWithGivenMembers: testMembers];
	NSDictionary* results = [self singleTestWithCrowd: testCrowd spiesViews: REVERSE
								chainBuilder: MWCB chainResolver: JOINCR acceptingAt: 4];
	int qs = [[results objectForKey: @"numberOfQuestions"] intValue];
    BOOL gotItRight = [[results objectForKey: @"gotItRight"] boolValue];
	NSLog(@"Number of questions = %i, gotItRight = %i",qs,gotItRight);
	[testCrowd showMembersFully];
}

-(void) paperTest
{
    NSMutableArray* testMembers = [[NSMutableArray alloc] init];
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //1
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //2
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //3
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //4
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //5
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //6
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //7
	[testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //8 should be a SPY but simpler to pretend is KNIGHT for views
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //9 
    [testMembers addObject: [NSNumber numberWithInt: SPY]];	//10
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //11
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]]; //12
    [testMembers addObject: [NSNumber numberWithInt: SPY]]; //13
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: SPY]];
    [testMembers addObject: [NSNumber numberWithInt: SPY]];
    [testMembers addObject: [NSNumber numberWithInt: SPY]];
	[testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    [testMembers addObject: [NSNumber numberWithInt: KNIGHT]];
    Crowd* testCrowd = [[Crowd alloc] initWithGivenMembers: testMembers];
	NSDictionary* results = [self singleTestWithCrowd: testCrowd spiesViews: REVERSE
								chainBuilder: MWCB chainResolver: NAIVECR acceptingAt: 10];
	int qs = [[results objectForKey: @"numberOfQuestions"] intValue];
    BOOL gotItRight = [[results objectForKey: @"gotItRight"] boolValue];
	NSLog(@"Number of questions = %i, gotItRight = %c",qs,gotItRight);
	[testCrowd showMembers];
    //won't work at moment
}

-(void)	findAverage: (int) L endAt: (int) M step: (int) k spiesViews: (int) spiesViews runs: (int) N
		chainBuilder: (int) cb chainResolver: (int) cr propSpies: (float) prop acceptingAtProp: (float) propAcc 
{
//finds the average number of questions needed for crowds with propSpies prop: so e.g. 
//if prop is 1/2 then get m-1 m+1 crowds
	int m; for (m=L;m<=M;m+=k) {
	    int spies = m*prop; int knights = m - spies;
		if (spies == -1) {spies = 0; knights -=1; } //hack to get it dead on if prop = 0;
		if (spies == knights) {spies -=1; knights +=1;} //and if prop = 1/2
		int l = m*propAcc; 
		if (propAcc == 0.5) { l -=1; };
		int totalQuestions = 0;
		//NSLog(@"About to conduct a run of tests with %i knights, %i spies",knights, spies);
		int i; for (i=0;i<N;i++) {
			if (i % 100 == 0) {
				NSLog(@"Iteration: %i",i);
			}
			NSDictionary* results = [self singleTestWithKnights: knights spies: spies spiesViews: spiesViews
												   chainBuilder: cb chainResolver: cr acceptingAt: l];
			int questionsAsked = [[results objectForKey: @"numberOfQuestions"] intValue];
			//int questionsToKnight = [[results objectForKey: @"questionsToKnight"] intValue];
			BOOL gotItRight = [[results objectForKey: @"gotItRight"] boolValue];
			if (!gotItRight) {
				NSLog(@"Stopping tests");
				break;
			}
			else {
				totalQuestions += questionsAsked;
				//NSLog(@"totalQuestions = %i",totalQuestions);
				//if ((i > 0) && (i % 10 == 0)) { NSLog(@"Iteration: %i",i); }
			}
        }
		float meanQuestions = totalQuestions;
		meanQuestions = meanQuestions / N;
		//NSLog(@"meanQuestions used = %f",meanQuestions);
		NSLog(@"%i, %.2f",m,meanQuestions);
    }
}

@end
